
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h>
#include <libxml/xmlmemory.h>

#include "vuc-glade-plugin.h"


#ifndef P_
#  define P_(x)		x
#endif

enum {
  VUC_CODE_0,
  VUC_CODE_SNIPPET_BEG,
  VUC_CODE_SNIPPET_END,
  VUC_CODE_SNIPPET_DATA,
  VUC_CODE_SNIPPET_WIDGET,
  VUC_CODE_LAUNCH
};

typedef struct _vuc_code_out vuc_code_out;

typedef void (* vuc_code_gen)(vuc_code_out * code, gint snippet, ...);

struct _vuc_code_out {
  GladeProject * project;
  gchar * name;
  vuc_code_gen gen;
  gchar * indent;
  gchar * cmt_beg, * cmt_end;

  gint put_cnt, get_cnt, rcv_cnt;
  GString * header;
  GString * body1, * body2, * body3, * main1, * main2, * data;
};


const gchar * vuc_gp2str (GladeProperty * gp);
GString * vuc_g_string_vnewf (const gchar * fmt, va_list args);
GString * vuc_g_string_newf (const gchar * fmt, ...);
gchar * vuc_strip_ext (gchar * name);

void vuc_file_wrf (const gchar * name, const gchar * fmt, ...);

gchar * vuc_camel2ul (gchar * name, gboolean upper);

gint vuc_run_with_msg (gchar * cmd, gint flags, GtkWidget * parent);
void vuc_null_childsetup (gpointer user_data);
gboolean vuc_run_async (gchar ** argv, gchar ** envp, GSpawnFlags flags);

void vuc_do_launch (const gchar * name, const gchar * args, GtkWidget * parent);

void vuc_code_c (vuc_code_out * code, gint snippet, ...);

void vuc_insert_toolbar_button (GtkToolbar * toolbar, gchar * label, gchar * icon, gchar * subdir, gint pos, GtkSignalFunc func, gpointer user_data);
void vuc_widget_tree_save (vuc_code_out * code, GladeWidget * widget, gint rec);

void vuc_on_prefs (GtkToolButton * toolbutton, gpointer user_data);
void vuc_on_create (GtkToolButton * toolbutton, gpointer user_data);
void vuc_on_launch (GtkToolButton * toolbutton, gpointer user_data);

void glade_plugin_toolbar_add (GladeWindow * window, GtkToolbar * toolbar);


const gchar * vuc_gp2str (GladeProperty * gp)
{
  if (gp)
    return g_value_get_string(gp->value);
  return NULL;
}

void vuc_widget_tree_save (vuc_code_out * code, GladeWidget * widget, gint rec)
{
//  gint i;
  const GList * kids, * props;

//  for (i=0; i<rec; i++) printf("  ");
//  printf("name:'%s'  adaptor->name:'%s'\n", widget->name, widget->adaptor->name);

  if (strncmp("Vuc", widget->adaptor->name, 3)==0) {
    GladeProperty * addr= NULL;
    GladeProperty * size= NULL;
    GladeProperty * init= NULL;
    GladeProperty * run= NULL;
    GladeProperty * decl= NULL;

    props= widget->properties;
    for ( ; props; props= props->next) {
      GladeProperty * p= props->data;
//      for (i=0; i<rec; i++) printf("  ");
//      printf("  prop id: '%s' name: '%s'\n", p->klass->id, p->klass->name);
      if (strcmp(p->klass->id, "vuc-addr")==0)
        addr= p;
      else
      if (strcmp(p->klass->id, "vuc-size")==0)
        size= p;
      else
      if (strcmp(p->klass->id, "vuc-init")==0)
        init= p;
      else
      if (strcmp(p->klass->id, "vuc-run")==0)
        run= p;
      else
      if (strcmp(p->klass->id, "vuc-decl")==0)
        decl= p;
    }

//    if (addr && size)
    if (init || run)
      code->gen(code, VUC_CODE_SNIPPET_WIDGET,
        widget->adaptor->name, widget->name,
        vuc_gp2str(init), vuc_gp2str(run), vuc_gp2str(decl)
//        g_value_get_int(addr->value), g_value_get_int(size->value)
      );
    else
      printf("Properties 'vuc-init' and 'vuc-run' must both be present.\n");
  }

  if (!(kids= glade_widget_adaptor_get_children(widget->adaptor, widget->object)))
    return;

  for ( ; kids; kids= kids->next) {
    GladeWidget * child= glade_widget_get_from_gobject(kids->data);
    if (child)
      vuc_widget_tree_save(code, child, rec+1);
  }
}


gchar * vuc_camel2ul (gchar * name, gboolean upper)
{
  gchar * buf= g_alloca(strlen(name)*2+1);
  gchar * p, * o= buf;

  for (p= name; *p; p++) {
    if (g_ascii_isupper(*p) && (p==name || !g_ascii_isupper(*(p-1))))
      *o++= '_';
    *o++= upper ? g_ascii_toupper(*p) : g_ascii_tolower(*p);
  }
  *o= 0;
  return g_strdup(&buf[1]);
}

GString * vuc_g_string_vnewf (const gchar * fmt, va_list args)
{
  GString * str= g_string_new(NULL);

  g_string_vprintf(str, fmt, args);
  return str;
}

GString * vuc_g_string_newf (const gchar * fmt, ...)
{
  va_list args;
  GString * str;

  va_start(args, fmt);
  str= vuc_g_string_vnewf(fmt, args);
  va_end(args);
  return str;
}

void vuc_file_wrf (const gchar * name, const gchar * fmt, ...)
{
  va_list args;
  GString * str= g_string_new(NULL);

  va_start(args, fmt);
  str= vuc_g_string_vnewf(fmt, args);
  g_file_set_contents(name, str->str, str->len, NULL);
  g_string_free(str, TRUE);
  va_end(args);
}

gchar * vuc_strip_ext (gchar * name)
{
  gchar * p= g_strrstr(name, ".");

  if (p) {
    gchar * o, c= *p;
    *p= 0; o= g_strdup(name); *p= c;
    return o;
  }
  return g_strdup(name);
}


void vuc_null_childsetup (gpointer user_data)
{
}

gboolean vuc_run_async (gchar ** argv, gchar ** envp, GSpawnFlags flags)
{
//  gchar ** envp= g_alloca(sizeof(gchar *));
  const gchar * dir= g_get_current_dir();
  GPid child_pid;

//  envp[0]= g_strdup_printf("PATH=%s", g_getenv("PATH"));
//  envp[0]= NULL;

  gboolean rc= g_spawn_async_with_pipes(
    dir, argv, envp, flags,
    vuc_null_childsetup, NULL,
    &child_pid, NULL, NULL, NULL, NULL
  );

  g_free((void *)dir);
  return rc;
}

gint vuc_run_with_msg (gchar * cmd, gint flags, GtkWidget * parent)
{
  GString * msg= g_string_new(NULL);
  gchar * std_out= NULL, * std_err= NULL;
  gboolean spawn_ok;
  gint exit_code;

  spawn_ok= g_spawn_command_line_sync(cmd, &std_out, &std_err, &exit_code, NULL);

  if (!spawn_ok || exit_code)
    g_string_append_printf(msg, "Exit code: %d\n", exit_code);
  if (std_out && *std_out)
    g_string_append_printf(msg, "\nstdout:\n----------\n%s----------\n", std_out);
  if (std_err && *std_err)
    g_string_append_printf(msg, "\nstderr:\n----------\n%s----------\n", std_err);

  if (!spawn_ok || exit_code || (std_err && *std_err) || (flags&1 && std_out && *std_out)) {
    GtkWidget * label, * dialog, * content_area;
    GString * title= vuc_g_string_newf(_("'%s' said:"), cmd);

    label= gtk_label_new(msg->str);
    dialog= gtk_dialog_new_with_buttons(title->str,
				      GTK_WINDOW(gtk_widget_get_toplevel(parent)),
				      GTK_DIALOG_MODAL,
				      GTK_STOCK_CLOSE,
				      GTK_RESPONSE_CLOSE,
				      NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_CLOSE);
    gtk_window_set_default_size(GTK_WINDOW(dialog), 300, 200);
//    gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 10);
    content_area= gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(content_area), label, TRUE, TRUE, 0);
    gtk_widget_show(label);

    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
    g_string_free(title, TRUE);
  }

  g_free(std_out);
  g_free(std_err);
  g_string_free(msg, TRUE);
  return exit_code;
}

void vuc_do_launch (const gchar * name, const gchar * args, GtkWidget * parent)
{
  gint exit_code;

  g_chdir(name);

//gchar * d= g_get_current_dir(); printf("  dir: '%s'  name: '%s'  args: '%s'\n", d, name, args); g_free(d);

  exit_code= vuc_run_with_msg("make", 0, parent);
  if (!exit_code) {
    gchar ** argv= g_alloca(3*sizeof(gchar *));
    argv[0]= g_strdup_printf("./%s", name);
    argv[1]= (void *)args;
    argv[2]= NULL;
    vuc_run_async(argv, NULL, G_SPAWN_SEARCH_PATH);
  }

  g_chdir("..");
}

void vuc_code_c (vuc_code_out * code, gint snippet, ...)
{
  va_list args;

  va_start(args, snippet);
  switch (snippet) {
    case VUC_CODE_SNIPPET_BEG:
      g_string_append_printf(
        code->body1,
        "#include <config.h>\n\n#include <stdio.h>\n#include <stdlib.h>\n#include <stdarg.h>\n#include <string.h>\n\n"
        "#include <gtk/gtk.h>\n#include <gdk/gdk.h>\n\n"
        "#include <vuc/libvuc.h>\n\n"
        "#include <%s.h>\n\n"
//        "#include \"main.h\"\n\n"
        "gchar * vuc_xml ();\n",
        code->name
      );
//      g_string_append_printf(code->body, "gchar * vuc_xml ();\n\n");
      g_string_append_printf(
        code->main1,
        "int main (int argc, char * argv[])\n"
        "{\n"
        "%sint rc= 0;\n"
        "%sGtkBuilder * builder;\n",
        code->indent, code->indent
      );
      g_string_append_printf(
        code->main2,
        "%sbuilder= vuc_init(BAUDRATE, -1, vuc_xml(), argc, argv);\n",
        code->indent
      );
      break;

    case VUC_CODE_SNIPPET_DATA: {
      gchar * p, * txt= va_arg(args, char *);
      g_string_append_printf
        (code->data, "gchar * vuc_xml (void)\n{\n%sreturn\n\"", code->indent);
      for (p= txt; *p; p++) {
        if (*p=='"')
          code->data= g_string_append(code->data, "\\\"");
        else
        if (*p=='\n')
          code->data= g_string_append(code->data, "\\n\"\n\"");
        else
          code->data= g_string_append_c(code->data, *p);
      }
      g_string_append_printf(code->data, "\";\n}\n");
      break;
    }

    case VUC_CODE_SNIPPET_WIDGET: {
      gchar * type= va_arg(args, char *);
      gchar * name= va_arg(args, char *);
      gchar * init= va_arg(args, char *);
      gchar * run= va_arg(args, char *);
      gchar * decl= NULL;
      gchar * type_l= vuc_camel2ul(type, 0);
      gchar * type_u= vuc_camel2ul(type, 1);

      if (strcmp("VucObject", type)==0) {
        decl= va_arg(args, char *);
        if (strlen(decl)==0)
          decl= NULL;
        else
          g_string_append_printf(
            code->body2,
            "#define %s_(var) ((%s_x *)%s->xdata)->var\n"
            "typedef struct\n"
            "{\n"
            "%s;\n"
            "} %s_x;\n",
            name, name, name, decl, name
          );
      }

      if (run && strlen(run)) {
        g_string_append_printf(
          code->body1,
          "void %s_run (%s * obj, gpointer udata);\n",
          name, type
        );
        g_string_append_printf(
          code->body3,
          "void %s_run (%s * obj, gpointer udata)\n"
          "{\n"
          "gint addr= obj->vuc_addr, size= obj->vuc_size;\n",
          name, type
        );

        if (decl && strcmp("VucObject", type)==0)
          g_string_append_printf(code->body3, "%s_x * data= (%s_x *)obj->xdata;\n", name, name);
        else
        if (strcmp("VucButton", type)==0)
          g_string_append_printf(code->body3, "gboolean val= vuc_button_get_val(obj);\n");
        else
        if (strcmp("VucHRadio", type)==0 || strcmp("VucVRadio", type)==0)
          g_string_append_printf(code->body3, "gint val= %s_get_val(obj);\n", type_l);
        else
        if (strcmp("VucHFader", type)==0 || strcmp("VucVFader", type)==0)
          g_string_append_printf(code->body3, "gdouble val= %s_get_val(obj);\n", type_l);

        g_string_append_printf
          (code->body3, "\n%s;\n}\n\n", run);
      }

      g_string_append_printf(code->body2, "%s * %s;\n\n", type, name);

      g_string_append_printf(
        code->main2,
        "\n%s%s= %s(gtk_builder_get_object(builder, \"%s\"));\n",
        code->indent, name, type_u, name
      );
      if (decl && strcmp("VucObject", type)==0)
        g_string_append_printf(
          code->main2,
          "%s%s->xdata= g_malloc0(sizeof(%s_x));\n",
          code->indent, name, name
        );

      if (init && strlen(init)) {
//        g_string_append_printf(code->main1, "%s%s * %s;\n", code->indent, type, name);
        g_string_append_printf(
          code->main2,
          "{\n"
          "%s * obj= %s; gint addr= obj->vuc_addr, size= obj->vuc_size;\n",
          type, name
        );
        if (decl && strcmp("VucObject", type)==0)
          g_string_append_printf
            (code->main2, "%s_x * data= obj->xdata;\n", name);
        g_string_append_printf
          (code->main2, "\n%s;\n}\n", init);
      }

//      if (strcmp("VucScope", type)==0) {
//        code->rcv_cnt++;
//      } else
      if (run) {
        gchar * signal= NULL;

//        code->put_cnt++;
        if (strcmp("VucButton", type)==0)
          signal= "toggled";
        else
        if (strcmp("VucHFader", type)==0 || strcmp("VucVFader", type)==0)
          signal= "value-changed";

        if (signal)
          g_string_append_printf(
            code->main2,
            "%sgtk_signal_connect(GTK_OBJECT(%s), \"%s\", GTK_SIGNAL_FUNC(%s_run), %s);\n",
            code->indent, name, signal, name, name
          );
      }

      g_free(type_u);
      break;
    }

    case VUC_CODE_SNIPPET_END: {
      GString * fn;

      fn= vuc_g_string_newf("%s.h", code->name);
      vuc_file_wrf(fn->str, code->header->str);
      g_string_free(fn, TRUE);

      g_mkdir(code->name, 0775);
      g_chdir(code->name);

      fn= vuc_g_string_newf("%s.c", code->name);
      g_string_append_printf(
        code->main2,
        "\n"
        "%sif (!rc) rc= vuc_go();\n"
        "\n"
        "%sreturn rc;\n"
        "}\n",
        code->indent, code->indent
      );
      vuc_file_wrf(
        fn->str,
        "%s\n\n%s\n%s\n%s\n%s\n%s\n",
        code->body1->str, code->body2->str, code->body3->str, code->main1->str, code->main2->str, code->data->str
      );
      g_string_free(fn, TRUE);

      vuc_file_wrf("Makefile", 
"DEBUG = -g -O0 -save-temps\n"
"\n"
"%s:	%s.o\n"
"	export LD_RUN_PATH=$$(expr substr \"$$(pkg-config --libs-only-L vuc)\" 3 9999); \\\n"
"	$(CC) -o $@ $< $$(pkg-config --libs gtk+-2.0) $$(pkg-config --libs vuc) $(DEBUG)\n"
"\n"
"%%.o:	%%.c ../%%.h\n"
"	$(CC) -o $@ -c $< -I.. $$(pkg-config --cflags gtk+-2.0) $$(pkg-config --cflags vuc) $(DEBUG)\n"
"\n"
"clean:\n"
"	rm -f %s.[ios]\n",
        code->name, code->name, code->name
      );

      g_chdir("..");

      break;
    }

    case VUC_CODE_LAUNCH: {
      GtkWidget * parent= va_arg(args, void *);
      VucGladeProjectPlugin * data= vuc_glade_project_get_data(code->project);

      vuc_do_launch(code->name, gtk_entry_buffer_get_text(data->exe_args), parent);
      break;
    }

  }
  va_end(args);
}


static gboolean vuc_glade_gen_xml (GladeProject * project, int * len, char ** txt)
{
  GladeXmlContext * context;
  GladeXmlDoc * doc;

  g_return_val_if_fail(GLADE_IS_PROJECT(project), FALSE);

  if (glade_project_is_loading(project))
    return FALSE;

//  if (!glade_project_verify(project, TRUE))
//    return FALSE;

  context= glade_project_write(project);
  doc= glade_xml_context_get_doc(context);
  xmlKeepBlanksDefault(0);
  xmlDocDumpFormatMemoryEnc((xmlDocPtr)doc, (xmlChar **)txt, len, "UTF-8", 1);
  glade_xml_context_destroy(context);

  return TRUE;
}


void vuc_on_prefs (GtkToolButton * toolbutton, gpointer user_data)
{
  GladeWindow * window= user_data;
  GladeProject * project = glade_design_view_get_project(window->priv->active_view);

  glade_project_preferences(project);
}


void vuc_on_create (GtkToolButton * toolbutton, gpointer user_data)
{
  GladeWindow * window= user_data;
  GladeProject * project= glade_design_view_get_project(window->priv->active_view);
  const GList * list;
  vuc_code_out * code;
  gchar * name= glade_project_get_name(project);
  int xml_len;
  char * xml_txt;
  VucGladeProjectPlugin * data= project->plugin;

  if (!glade_project_get_modified(project))
    save_cb(NULL, window); // ask for filename if new and save

  if (!vuc_glade_gen_xml(project, &xml_len, &xml_txt))
    return;

  code= g_malloc0(sizeof(vuc_code_out));
  code->project= project;
  code->name= vuc_strip_ext(name); g_free(name);
  code->indent= "  ";
  code->gen= vuc_code_c;
  code->header= g_string_new(vuc_gtk_text_buffer_get_text(data->header));
  code->body1= g_string_new(NULL);
  code->body2= g_string_new(NULL);
  code->body3= g_string_new(NULL);
  code->main1= g_string_new(NULL);
  code->main2= g_string_new(NULL);
  code->data= g_string_new(NULL);
  code->gen(code, VUC_CODE_SNIPPET_BEG);
  code->gen(code, VUC_CODE_SNIPPET_DATA, xml_txt, xml_len);
  xmlFree(xml_txt);

  glade_project_get_path(project);
  for (list= glade_project_get_objects(project); list; list= list->next) {
    GladeWidget * widget;

    widget= glade_widget_get_from_gobject(list->data);
    if (widget->parent == NULL)
      vuc_widget_tree_save(code, widget, 0);
  }

  code->gen(code, VUC_CODE_SNIPPET_END);

  g_free(code->name);
  g_string_free(code->header, TRUE);
  g_string_free(code->body1, TRUE);
  g_string_free(code->body2, TRUE);
  g_string_free(code->body3, TRUE);
  g_string_free(code->main1, TRUE);
  g_string_free(code->main2, TRUE);
  g_string_free(code->data, TRUE);
}


void vuc_on_launch (GtkToolButton * toolbutton, gpointer user_data)
{
  GladeWindow * window= user_data;
  GladeProject * project= glade_design_view_get_project(window->priv->active_view);
  gchar * name= glade_project_get_name(project);
  vuc_code_out * code= g_malloc0(sizeof(vuc_code_out));

  code->project= project;
  code->name= vuc_strip_ext(name); g_free(name);
  vuc_code_c(code, VUC_CODE_LAUNCH, toolbutton);

  g_free(code->name);
  g_free(code);
}


void vuc_insert_toolbar_button (GtkToolbar * toolbar, gchar * label, gchar * icon, gchar * subdir, gint pos, GtkSignalFunc func, gpointer user_data)
{
  GtkWidget * ic;
  GtkToolItem * it;

  ic= glade_utils_get_icon(icon, subdir, TRUE);
  gtk_widget_show(GTK_WIDGET(ic));
  it= gtk_tool_button_new(ic, label);
  gtk_widget_show(GTK_WIDGET(it));
  gtk_toolbar_insert(toolbar, it, pos);
  gtk_signal_connect(GTK_OBJECT(it), "clicked", func, user_data);
}

void glade_plugin_toolbar_add (GladeWindow * window, GtkToolbar * toolbar)
{
  GtkToolItem * sep= gtk_separator_tool_item_new();
  gchar * subdir= g_build_filename("hicolor", "22x22", "actions", NULL);

  gtk_widget_show(GTK_WIDGET(sep));
  vuc_insert_toolbar_button(toolbar, P_("Prefs"), "vuc-btn-prefs.png", subdir, 4, GTK_SIGNAL_FUNC(vuc_on_prefs), window);
  vuc_insert_toolbar_button(toolbar, P_("Create"), "vuc-btn-create.png", subdir, 5, GTK_SIGNAL_FUNC(vuc_on_create), window);
  vuc_insert_toolbar_button(toolbar, P_("Launch"), "vuc-btn-launch.png", subdir, 6, GTK_SIGNAL_FUNC(vuc_on_launch), window);
  gtk_toolbar_insert(toolbar, sep, 7);
  g_free(subdir);
}
